/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef __MPM_TRANSPORT_H__
#define __MPM_TRANSPORT_H__

#include <stdint.h>
#include <stdbool.h>
#include <fcntl.h>
#include <sys/mman.h>

#define DEFAULT_CONFIG_FILE_NAME "/etc/mpm/mpm_config.json"

#define mpm_transport_h void *
#define mpm_transport_trans_h void *

typedef struct mpm_transport_open_tag {
	mode_t open_mode;
	uint32_t msec_timeout;
} mpm_transport_open_t;

typedef struct mpm_transport_read_tag {
	uint32_t dummey;
} mpm_transport_read_t;

typedef struct mpm_transport_write_tag {
	uint32_t dummey;
} mpm_transport_write_t;

typedef struct mpm_transport_mmap_tag {
	int mmap_prot;
	int mmap_flags;
} mpm_transport_mmap_t;

/*
	slave_name:  should match the name in config file.
	cfg: runtime configuration needed for this device.
	In case of shared memory it might not have anything.
	Returns mpm_transport_handle or NULL on error.
*/
mpm_transport_h mpm_transport_open(char *slave_name, mpm_transport_open_t *cfg);

/*
	Close mpm_transport connection.
*/
void mpm_transport_close(mpm_transport_h h);

/*
	Read from slave[DSP] local address and length into the buffer.
	Returns 0 on success, error code on error.
*/
int mpm_transport_read(mpm_transport_h h, uint32_t addr, uint32_t length, char *buf, mpm_transport_read_t *cfg);

/*
	Write to slave[DSP] local address "addr" with "length" from the buffer "buf".
	Returns 0 on success, error code on error.
*/
int mpm_transport_write(mpm_transport_h h, uint32_t addr, uint32_t length, char *buf, mpm_transport_write_t *cfg);

/*
	Creates a mmap for the memory location and returns virtual address.
*/
void *mpm_transport_mmap(mpm_transport_h h, uint32_t addr, uint32_t length, mpm_transport_mmap_t *cfg);

/*
	Unmaps transport mmap
*/
void mpm_transport_munmap(mpm_transport_h h, void *va, uint32_t length);

/*
	Similar to mpm_transport_read, except memcpy is replaced with EDMA3 operation.
	Returns transaction handle.
*/
mpm_transport_trans_h mpm_transport_get_initiate(mpm_transport_h h, uint32_t from_addr, uint32_t to_addr, uint32_t length, bool is_blocking, mpm_transport_read_t *cfg);

/*
	Similar to mpm_transport_write, except memcpy is replaced with EDMA3 operation.
	Returns transaction handle.
*/
mpm_transport_trans_h mpm_transport_put_initiate(mpm_transport_h h, uint32_t to_addr, uint32_t from_addr, uint32_t length, bool is_blocking, mpm_transport_write_t *cfg);

/*
	Similar to mpm_transport_get_initiate(), but with multiple transfers linked over EDMA3.
	Returns transaction handle.
*/
mpm_transport_trans_h mpm_transport_get_initiate_linked(mpm_transport_h h, uint32_t num_links, uint32_t from_addr[], uint32_t to_addr[], uint32_t length[], bool is_blocking, mpm_transport_write_t *cfg);

/*
	Similar to mpm_transport_put_initiate(), but with multiple transfers linked over EDMA3.
	Returns transaction handle.
*/
mpm_transport_trans_h mpm_transport_put_initiate_linked(mpm_transport_h h, uint32_t num_links, uint32_t to_addr[], uint32_t from_addr[], uint32_t length[], bool is_blocking, mpm_transport_write_t *cfg);

/*
	Checks for EDMA completion. Clears the pending interrupt if complete.
	Returns:
	   -1 = error
		0 = In progress
		1 = complete
*/
int mpm_transport_transfer_check(mpm_transport_h h, mpm_transport_trans_h th);

/*
	Creates a mmap for the memory location and returns physical address.
*/
uint32_t mpm_transport_phys_map(mpm_transport_h h, uint32_t addr, uint32_t length, mpm_transport_mmap_t *cfg);

/*
	Long address version of mpm_transport_get_initiate()
	Returns transaction handle.
*/
mpm_transport_trans_h mpm_transport_get_initiate64(mpm_transport_h h, uint64_t from_addr, uint64_t to_addr, uint32_t length, bool is_blocking, mpm_transport_read_t *cfg);

/*
	Long address version of mpm_transport_put_initiate()
	Returns transaction handle.
*/
mpm_transport_trans_h mpm_transport_put_initiate64(mpm_transport_h h, uint64_t to_addr, uint64_t from_addr, uint32_t length, bool is_blocking, mpm_transport_write_t *cfg);

/*
	Long address version of mpm_transport_mmap()
	Creates a mmap for the memory location and returns virtual address.
*/
uint32_t mpm_transport_phys_map64(mpm_transport_h h, uint64_t addr, uint32_t length, mpm_transport_mmap_t *cfg);

#endif
